//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <except.h>
#pragma hdrstop

#include "FileMan.h"
#include "Mixmaster.h"
#include "zlib.h"
#include "view.h"

const int MaxNest = 16;
const char *FailedFileManagerInit = "I was unable to initialize the file manager";
const char *FailedFileManagerIndex = "I was unable to initialize the file manager master index";
const char *FailedFileManagerSubIndex = "I was unable to initialize the file manager index ";
const char *FailedToRetrieveData = "I was unable to load the requested data: ";
const char *FailedToRenameIndex = "I was unable to complete the rename operation";
const char *FailedToRenameItem = "I was unable to complete the rename operation";
//---------------------------------------------------------------------------
#pragma link "zlib.lib"
bool DBFATALerror = false;
bool UseCompression = true;
//---------------------------------------------------------------------------
int __fastcall PutAppBinary(AnsiString &path, TDataBuf *buf, AnsiString &datafile)
{
   if (DBFATALerror) return 0;
   AnsiString nodetext;
   int i,status=DB_OK;
   TFMindex *idx[MaxNest]={ 0 };

   TFileManager *fman = new TFileManager(datafile);
   if (fman->IsError(FailedFileManagerInit))
      goto exit;

   idx[0] = new TFMindex(fman);
   if (fman->IsError(FailedFileManagerIndex))
      goto exit;

   i=1; while ((nodetext=GetParam(i-1,path,PS.c_str()))!=NULL)
   {
      if (GetParam(i,path,PS.c_str()).IsEmpty())
      {
         idx[i-1]->PutBinary(nodetext,buf);
         if (fman->IsError(FailedToRetrieveData+nodetext))
            goto exit;

         status = fman->lostatus;
         break;
      }
      idx[i] = new TFMindex(fman,nodetext,idx[i-1]); i++;
      if (fman->IsError(FailedFileManagerSubIndex+nodetext))
         goto exit;
   }

exit:
   for (int x=MaxNest-1; x>0; x--)
      if (idx[x]) delete idx[x];
   delete fman;
   return status;
}
//---------------------------------------------------------------------------
int __fastcall PutAppBinaryData(AnsiString &path, char *data, int length, AnsiString &datafile)
{
   if (DBFATALerror) return 0;
   TDataBuf *buffer = new TDataBuf(data,length);
   int status = PutAppBinary(path,buffer);
   delete buffer;
   return status;
}
//---------------------------------------------------------------------------
int __fastcall GetAppBinary(AnsiString &path, TDataBuf *buf, AnsiString &datafile)
{
   int i,status=DB_OK;
   if (DBFATALerror) return 0;
   if (!FileExists(datafile)) return status;

   AnsiString nodetext;
   TFMindex *idx[MaxNest] = { 0 };

   TFileManager *fman = new TFileManager(datafile);
   if (fman->IsError(FailedFileManagerInit))
      goto exit;

   idx[0] = new TFMindex(fman);
   if (fman->IsError(FailedFileManagerIndex))
      goto exit;

   i=1; while ((nodetext=GetParam(i-1,path,PS.c_str()))!=NULL)
   {
      if (GetParam(i,path,PS.c_str()).IsEmpty())
      {
         if (idx[i-1]->GetSize(nodetext)>0)
         {
            idx[i-1]->GetBinary(nodetext,buf);
            if (fman->IsError(FailedToRetrieveData+nodetext))
               goto exit;
         }
         status = fman->lostatus;
         break;
      }
      idx[i] = new TFMindex(fman,nodetext,idx[i-1]); i++;
      if (fman->IsError(FailedFileManagerSubIndex+nodetext))
         goto exit;
   }

exit:
   for (int x=MaxNest-1; x>0; x--)
      if (idx[x]) delete idx[x];
   delete fman;
   return status;
}
//---------------------------------------------------------------------------
int __fastcall GetAppDataMainIndex(TStrings *items, AnsiString &datafile)
{
   if (DBFATALerror) return 0;
   TFMindex *idx=NULL;

   TFileManager *fman = new TFileManager(datafile);
   if (fman->IsError(FailedFileManagerInit))
      goto exit;

   idx = new TFMindex(fman);
   if (fman->IsError(FailedFileManagerIndex))
      goto exit;

   items->AddStrings(idx);
exit:
   int status = fman->lostatus;
   if (idx) delete idx;
   delete fman;
   return status;
}
//---------------------------------------------------------------------------
int __fastcall GetAppData(TStrings *items, AnsiString &path, AnsiString &datafile)
{
   if (DBFATALerror) return 0;
   items->Clear();
   int status = DB_OK;
   TDataBuf *buf = new TDataBuf();
   if (buf!=NULL)
   {
      status = GetAppBinary(path,buf,datafile);
      items->Text = s(buf->data);
      delete buf;
   }
   return status;
}
//---------------------------------------------------------------------------
int __fastcall PutAppData(AnsiString &path, TStrings *items, AnsiString &datafile)
{
   if (DBFATALerror) return 0;
   TDataBuf *buf = NULL;
   if (items)
   {
      buf = new TDataBuf();
      if (buf->status==FB_OK && items->Count)
         buf->Set(items->Text.c_str(),items->Text.Length());
   }
   int ret = PutAppBinary(path,buf,datafile);
   if (buf) delete buf;
   return ret;
}
//---------------------------------------------------------------------------
bool __fastcall AppBinaryExists(AnsiString &path, TDataBuf *data, AnsiString &datafile)
{
   bool exists=false;
   if (DBFATALerror) return 0;
   TDataBuf *tmp = new TDataBuf();
   if (GetAppBinary(path,tmp,datafile))
   {
      if (data) data->Set(tmp->data,tmp->length);
      exists = tmp->length>0;
   }
   delete tmp;
   return exists;
}
//---------------------------------------------------------------------------
bool __fastcall AppDataExists(AnsiString &path, TStrings *data, AnsiString &datafile)
{
   bool exists=false;
   if (DBFATALerror) return 0;
   TStringList *tmp = new TStringList();
   if (GetAppData(tmp,path,datafile)==DB_OK)
   {
      if (data) data->Text = tmp->Text;
      exists=tmp->Count>0;
   }
   delete tmp;
   return exists;
}
//---------------------------------------------------------------------------
AnsiString __fastcall GetAppIni(AnsiString &key, AnsiString &defvalue, AnsiString &datafile)
{
   AnsiString val="";
   if (DBFATALerror) return 0;
   TStringList *ini = new TStringList();
   if (GetAppData(ini,AppIniValues,datafile)==DB_OK)
      val = ini->Values[key];
   delete ini;
   if (val.IsEmpty()) val = defvalue;
   return val;
}
//---------------------------------------------------------------------------
int __fastcall PutAppIni(AnsiString &key, AnsiString &Value, AnsiString &datafile)
{
   int ret=-1;
   if (DBFATALerror) return 0;
   TStringList *ini = new TStringList();
   if (GetAppData(ini,AppIniValues,datafile)==DB_OK)
   {
      ini->Values[key] = Value;
      ret = PutAppData(AppIniValues,ini,datafile);
   }
   delete ini;
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall RenameAppIndex(AnsiString &Path, AnsiString &Newname,
   AnsiString &datafile)
{
   if (DBFATALerror) return 0;
   String nodetext;
   int i,pos,status = DB_OK;
   TFMindex *idx[MaxNest] = { 0 };

   TFileManager *fman = new TFileManager(datafile);
   if (fman->IsError(FailedFileManagerInit))
      goto exit;

   idx[0] = new TFMindex(fman);
   if (fman->IsError(FailedFileManagerIndex))
      goto exit;

   i=1; while ((nodetext=GetParam(i-1,Path,PS.c_str()))!=NULL)
   {
      if (GetParam(i,Path,PS.c_str()).IsEmpty())
      {
         if ((pos=idx[i-1]->IndexOf(nodetext))>=0)
         {
            TObject *obj = idx[i-1]->Objects[pos];
            idx[i-1]->Strings[pos] = Newname;
            idx[i-1]->Objects[pos] = obj;
            idx[i-1]->Save();
            if (fman->IsError(FailedToRenameIndex+nodetext))
               goto exit;
         }
         status = fman->lostatus;
         break;
      }
      idx[i] = new TFMindex(fman,nodetext,idx[i-1]); i++;
      if (fman->IsError(FailedFileManagerSubIndex+nodetext))
         goto exit;
   }

exit:
   for (int x=MaxNest-1; x>0; x--)
      if (idx[x]) delete idx[x];
   delete fman;
   return status;
}
//---------------------------------------------------------------------------
int __fastcall RenameAppItem(AnsiString &Path, AnsiString &Newname,
   AnsiString &datafile)
{
   if (DBFATALerror) return 0;
   Path = Path.Trim(); //sm(Path);
   AnsiString nodetext;
   int i,pos,status=DB_OK;
   TFMindex *idx[MaxNest]={ 0 };

   TFileManager *fman = new TFileManager(datafile);
   if (fman->IsError(FailedFileManagerInit))
      goto exit;

   idx[0] = new TFMindex(fman);
   if (fman->IsError(FailedFileManagerIndex))
      goto exit;

   i=1; while ((nodetext=GetParam(i-1,Path,PS.c_str()))!=NULL)
   {
      if (GetParam(i,Path,PS.c_str()).IsEmpty())
      {
         pos = idx[i-1]->IndexOf(nodetext);
         if (pos>=0)
         {
            idx[i-1]->Strings[pos] = Newname;
            idx[i-1]->Save();
            if (fman->IsError(FailedToRenameItem+nodetext))
               goto exit;
         }
         status = fman->lostatus;
         break;
      }
      idx[i] = new TFMindex(fman,nodetext,idx[i-1]); i++;
      if (fman->IsError(FailedFileManagerSubIndex+nodetext))
         goto exit;
   }

 exit:
   for (int x=MaxNest-1; x>0; x--)
      if (idx[x]) delete idx[x];
   delete fman;
   return status;
}
//---------------------------------------------------------------------------
// TFileManager
//----------------------------------------------------------------------------
__fastcall TFileManager::TFileManager(const AnsiString &Filename, bool CreateIfNeed)
   : TObject()
{
   DBFATALerror = false;
   filename = Filename;
   stats = new TList();
   bool exists = FileExists(filename);
   if (!exists && !CreateIfNeed)
   {
      status(DB_FATALERR);
      status(DB_NODBERR);
      stream = NULL;
      return;
   }
   try
   {
      stream = new TFileStream(filename,(unsigned short)(exists?fmOpenReadWrite:fmCreate));
      delStream = true;
   }
   catch(...)
   {
      if (stream) delete stream;
      status(DB_FATALERR);
      status(DB_OPENDBEXC);
      return;
   }
   Init();
}
//----------------------------------------------------------------------------
__fastcall TFileManager::TFileManager(TStream *Stream) : TObject()
{
  delStream = false;
  stream = Stream;
  Init();
}
//---------------------------------------------------------------------------
__fastcall TFileManager::~TFileManager()
{
   delete stats;
   delete freeList;
   if (delStream) delete stream;
}
//---------------------------------------------------------------------------
void __fastcall TFileManager::Init()
{
  if (dbstatus!=DB_OK) return;
  startOfData = FLNIL; endOfData = strlen(FILEHEADER)+(2*OFSSIZE);
  freeList = new TFreeList();

  if(!stream->Size)  // Virgin file--must initialize...
  {
     if (!WriteHeader() || !WriteStartAndEnd() || !WriteFreeList())
        status(DB_INITVIRGINDBERR);
  }
  if (!ReadHeader() || !ReadStartAndEnd() || !ReadFreeList())
     status(DB_INITEXISTDBERR);

  DBFATALerror = FatalError();
}
//---------------------------------------------------------------------------
void __fastcall TFileManager::status(int statnum)
{
   if (statnum!=DB_OK) stats->Insert(0,(void*)statnum);
   else stats->Clear();
}
//---------------------------------------------------------------------------
int __fastcall TFileManager::HiStatus()
{
   if (stats->Count>0) return (int)stats->Items[0];
   return DB_OK;
}
//---------------------------------------------------------------------------
int __fastcall TFileManager::LoStatus()
{
   int c = stats->Count;
   if (c>0) return (int)stats->Items[c-1];
   return DB_OK;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::ReadHeader()
{
   status(DB_OK);
   int hdrlen = strlen(FILEHEADER);
   char *header = new char[hdrlen+1];
   try
   {
      if (!ReadOfs(0,header,hdrlen))
      {
         status(DB_FATALERR);
         status(DB_READHEADERR);
         delete [] header;
         return false;
      }
   }
   catch(...)
   {
      status(DB_FATALERR);
      status(DB_READHEADEXC);
      delete [] header;
      return false;
   }
   if (memcmp(header,FILEHEADER,hdrlen)!=0 &&
       memcmp(header,FILEHEADE1,hdrlen)!=0)
   {
      status(DB_FATALERR);
      status(DB_INVALIDHEADERR);
      delete [] header;
      return false;
   }

   delete [] header;
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::ReadStartAndEnd()
{
   status(DB_OK);
   try
   {
      if (!ReadOfs(strlen(FILEHEADER),&startOfData,2*OFSSIZE))
      {
         status(DB_FATALERR);
         status(DB_READSODERR);
         return false;
      }
   }
   catch(...)
   {
     status(DB_FATALERR);
     status(DB_READSODEXC);
     return false;
   }
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::ReadFreeList()
{
  status(DB_OK);
  try
  {
     if (!SeekTo(endOfData)) return false;
     if (!freeList->LoadFromStream(stream))
     {
        status(DB_FATALERR);
        status(DB_READERR);
        return false;
     }
  }
  catch(...)
  {
     status(DB_FATALERR);
     status(DB_READEXC);
     return false;
  }
  return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::WriteHeader()
{
   status(DB_OK);
   try
   {
      if (!WriteOfs(0,FILEHEADER,strlen(FILEHEADER)))
      {
         status(DB_FATALERR);
         status(DB_WRITEHEADERR);
         return false;
      }
   }
   catch(...)
   {
      status(DB_FATALERR);
      status(DB_WRITEHEADEXC);
      return false;
   }
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::WriteFreeList()
{
  status(DB_OK);
  try
  {
     if (!SeekTo(endOfData))
     {
        status(DB_FATALERR);
        status(DB_SEEKERR);
        return false;
     }
     if (!freeList->SaveToStream(stream))
     {
        status(DB_FATALERR);
        status(DB_WRITEERR);
        return false;
     }
  }
  catch(...)
  {
     status(DB_FATALERR);
     status(DB_WRITEEXC);
     return false;
  }
  return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::WriteStartAndEnd()
{
   status(DB_OK);
   try
   {
      if (!WriteOfs(strlen(FILEHEADER),&startOfData,2*OFSSIZE))
      {
         status(DB_FATALERR);
         status(DB_WRITEERR);
         return false;
      }
   }
   catch(...)
   {
     status(DB_FATALERR);
     status(DB_WRITEEXC);
     return false;
   }
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::SeekTo(TFLoffset offset)
{
   status(DB_OK);
   bool ok;
   try
   {
      if ((ok=stream->Seek(offset,soFromBeginning)==offset)==false)
         status(DB_SEEKERR);
   }
   catch(...)
   {
      status(DB_SEEKEXC);
      return false;
   }
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::WriteOfs(const TFLoffset ofs, void *data, long size)
{
   if (!SeekTo(ofs)) return false;
   return WriteDat(data,size);
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::WriteDat(void *dat, long size)
{
   status(DB_OK);
   bool ok;
   try
   {
      if ((ok=stream->Write(dat,size)==size)==false)
         status(DB_WRITEERR);
   }
   catch(...)
   {
      status(DB_WRITEEXC);
      return false;
   }
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::ReadOfs(const TFLoffset ofs, void *data, long size)
{
   if (!SeekTo(ofs)) return false;
   return ReadDat(data,size);
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::ReadDat(void *dat, long size)
{
   bool ok;
   status(DB_OK);
   try
   {
      if ((ok=stream->Read(dat,size)==size)==false)
         status(DB_READERR);
   }
   catch(...)
   {
      status(DB_READEXC);
      return false;
   }
   return ok;
}
//---------------------------------------------------------------------------
TFLoffset __fastcall TFileManager::Allocate(TFLspace spaceRequested)
{
  status(DB_OK);
  TFLoffset ofs;
  if (spaceRequested<0) spaceRequested==0;
  try
  {
     TFLspace space = spaceRequested+sizeofHeader;
     if (!AllocateInFreeList(ofs,space) && !AllocateAtEnd(ofs,space))
     {
         status(DB_ALLOCERR);
         Deallocate(ofs);
         return FLNIL;
     }
     if (startOfData==FLNIL) startOfData=ofs;
     if (!WriteStartAndEnd() || !WriteFreeList())
     {
         status(DB_ALLOCERR);
         Deallocate(ofs);
         return FLNIL;
     }
  }
  catch(...)
  {
     status(DB_ALLOCERR);
     return FLNIL;
  }
  return ofs;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::AllocateInFreeList(TFLoffset &offset, TFLspace spaceRequired)
{
   status(DB_OK);
   offset = FLNIL;
   TFLoffset startOfAlloc;
   try
   {
      if ((startOfAlloc=freeList->Allocate(spaceRequired))==FLNIL)
      {
        status(DB_ALLOCERR);
        return false;
     }

     if (!WriteOfs(startOfAlloc,&spaceRequired,SPCSIZE) || !WriteFreeList())
     {
        status(DB_ALLOCERR);
        return false;
     }
  }
  catch(...)
  {
     status(DB_ALLOCERR);
     return false;
  }
  offset = startOfAlloc+sizeofHeader;
  return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::AllocateAtEnd(TFLoffset &offset, TFLspace spaceRequired)
{
  status(DB_OK);
  offset = FLNIL;
  TFLoffset ofs = endOfData+sizeofHeader;
  try
  {
     // Record the space allocated in the header:
     if (!WriteOfs(endOfData,&spaceRequired,SPCSIZE))
     {
        status(DB_ALLOCERR);
        return false;
     }
     // Update and record new endOfData:
     endOfData += spaceRequired;
     if (!WriteStartAndEnd())
     {
        status(DB_ALLOCERR);
        return false;
     }
  }
  catch(...)
  {
     status(DB_ALLOCEXC);
     return false;
  }
  offset = ofs;
  return true;
}
//---------------------------------------------------------------------------
void __fastcall TFileManager::Deallocate(TFLoffset offset)
{
   status(DB_OK);
   // Can't deallocate first allocated block or FLNIL:
   if (offset==FLNIL || offset==start()) return;
   // Start of record, including record header:
   TFLoffset startOfRecord = offset-(TFLoffset)sizeofHeader;
   TFLspace spaceFreed;

   if (!ReadOfs(startOfRecord,&spaceFreed,SPCSIZE) ||
       !AddToFreeList(startOfRecord,spaceFreed) || !WriteFreeList())
   {
      status(DB_ALLOCERR);
   }
}
//---------------------------------------------------------------------------
// Add a free space to current free list block.
bool __fastcall TFileManager::AddToFreeList(TFLoffset offset, TFLspace space)
{
   status(DB_OK);
   if (!space) return true;

   // If last record is deleted, redefine endOfData.
   if (offset+(TFLoffset)space>=endOfData)
   {  // If last record removed
      endOfData = offset;
      if (!WriteStartAndEnd())
         status(DB_WRITEFREEERR);
   }
   else
   if (!freeList->EnterFreeSpace(offset, space))
      status(DB_WRITEFREEEXC);

   return !stats->Count;
}
//---------------------------------------------------------------------------
/*bool __fastcall TFileManager::Getstatus(int &statnum, AnsiString &statmsg)
{
   statnum = 0; statmsg = "";
   if (status->Count)
   {
      statnum = (int)status->Items[0];
      statmsg = status->Strings[0];
      status->Delete(0);
      return true;
   }
   return false;
}*/
//---------------------------------------------------------------------------
bool __fastcall TFileManager::IsError(AnsiString Errmsg)
{
   if (!stats->Count) return false;
   for (int i=0; i<stats->Count; i++)
      Errmsg += "\n"+GetErrorMessage((int)stats->Items[i]);
   MessageBox(NULL,Errmsg.c_str(),"File Manager Error",
      MB_OK|MB_TASKMODAL);
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFileManager::FatalError()
{
   bool fatal=false;
   for (int i=0; i<stats->Count; i++)
   {
      if ((int)stats->Items[i]==DB_FATALERR)
      {
         fatal = true;
         break;
      }
   }
   return fatal;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFileManager::GetErrorMessage(int ErrNo)
{
   switch (ErrNo)
   {
      case DB_OK: "All is well!"; break;

      case DB_FATALERR: return "It's fatal!";
      case DB_OPENDBEXC: return "Error opening database";
      case DB_INITVIRGINDBERR: return "Error creating database";
      case DB_INITEXISTDBERR: return "Error initializing database";
      case DB_INVALIDHEADERR: return "Invalid database header";

      case DB_NODBERR: return "File not found";
      case DB_SEEKERR:
      case DB_SEEKEXC: return "Data seek error";
      case DB_READERR:
      case DB_READEXC: return "Data read error";
      case DB_WRITEERR:
      case DB_WRITEEXC: return "Data write error";

      case DB_READHEADERR:
      case DB_READHEADEXC:
      case DB_WRITEHEADERR:
      case DB_WRITEHEADEXC: return "Database header error";

      case DB_READSODERR:
      case DB_READSODEXC:
      case DB_WRITESODERR:
      case DB_WRITESODEXC: return "Database start-of-data error";

      case DB_READFREEERR:
      case DB_READFREEEXC:
      case DB_WRITEFREEERR:
      case DB_WRITEFREEEXC:
      case DB_WRITEALLOCENDERR:
      case DB_WRITEALLOCSODERR:
      case DB_ALLOCFREEERR:
      case DB_ALLOCFREEEXC:
      case DB_ADDFREEWRITEERR:
      case DB_ADDFREEENTERERR:
      case DB_WRITEALLOCFREEERR: return "Freelist allocation error";
      case DB_DEALOCREADFREEERR:
      case DB_DEALLOCFREEERR: return "Freelist deallocation error";

      case DB_WRITEALLOCERR:
      case DB_ALLOCERR:
      case DB_ALLOCEXC:
      case DB_ALLOCATENDEXC: return "Data allocation error";
      case DB_DEALOCREADSORERR: return "Data de-allocation error";

      case DB_GETBINARYERR:
      case DB_GETBINALLOCEXC: return "Error loading program data from disk";
      case DB_PUTBINARYERR: return "Error saving program data to disk";

      case DB_OPENINDEXERR:
      case DB_OPENINDEXEXC:
      case DB_ALLOCINDEXEXC: return "Error opening database index";

      case DB_WRITEINDEXERR:
      case DB_WRITEINDEXOFSERR: return "Error storing database index";
      case DB_INITINDEXERR: "Error initializing database index";

      case DB_GETDATSIZEERR: return "Error getting data size";
      case DB_GETDATTYPEERR: return "Error getting data type";
      case DB_BADBUFERR: return "Initialization of data buffer failed";

      case DB_ZIPDEFLERR:
      case DB_ZIPINFLERR: return "Error processing program data";
      default: return s(ErrNo)+" Unknown file manager error!";
   }
   return "";
}
//---------------------------------------------------------------------------
// TDataBuf
//---------------------------------------------------------------------------
long __fastcall GetThisFileSize(const AnsiString FileName)
{
  TSearchRec SearchRec;
  if(FindFirst(ExpandFileName(FileName),faAnyFile,SearchRec)==0) return SearchRec.Size;
  else return 0;
}
//---------------------------------------------------------------------------
__fastcall TDataBuf::TDataBuf(const char *Data, TFLspace Length)
{
   status = DB_OK;
   sensitive = true;
   type[0]=type[1]=type[2]=type[3]='\0';
   data = NULL; length = orglength = 0;
   if (Data) Set(Data,Length);
}
//---------------------------------------------------------------------------
__fastcall TDataBuf::TDataBuf(const AnsiString &Data)
{
   status = DB_OK;
   sensitive = true;
   type[0]=type[1]=type[2]=type[3]='\0';
   data = NULL; length = orglength = 0;
   Set(Data.c_str(),Data.Length());
}
//---------------------------------------------------------------------------
__fastcall TDataBuf::~TDataBuf()
{
   if (data) Clear();
}
//---------------------------------------------------------------------------
void __fastcall TDataBuf::Clear()
{
   status = DB_OK;
   type[0]=type[1]=type[2]=type[3]='\0';
   try
   {
      if (data)
      {
         if (sensitive) memset(data,'\0',length);
         delete [] data;
         data = NULL;
      }
   }
   catch(...)
   {
      data = NULL;
   }
   orglength = 0;
   length = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDataBuf::Set(const char *Data, TFLspace Length)
// 'Set' completely reinitializes data structure
// 'SetData' below just set data and length
{
   Clear(); SetData(Data,Length);
   orglength = Length;
   sensitive = true;
}
//---------------------------------------------------------------------------
void __fastcall TDataBuf::SetData(const char *Data, TFLspace Length)
// see notes on 'Set', above. This does not change type[0], orglenth, sensitive
{
   status = DB_OK;
   try
   {
      if (data && length>=0)
      {
         if (sensitive) memset(data,'\0',length);
         delete [] data;
      }
      if ((data = new char[Length+1])==NULL)
      {
         status = FB_ALLOCBUFERR;
         orglength = 0;
         length = 0;
         return;
      }
      if (Data) memcpy(data,Data,Length); else memset(data,'\0',Length);
      data[Length] = '\0';
      length = Length;
   }
   catch(...)
   {
      status = FB_ALLOCBUFEXC;
      data = NULL;
      orglength = 0;
      length = 0;
      return;
   }
}
//---------------------------------------------------------------------------
bool __fastcall TDataBuf::Encrypted()
{
   return type[3]==ENCRYPTED;
}
//---------------------------------------------------------------------------
int __fastcall TDataBuf::Encrypt(const AnsiString &password)
{
   status = DB_OK;
   int result;
   try
   {
      if (type[3]!=CLEARTEXT) return 0;
      result=data_encrypt(&data,length,password.c_str(),password.Length());
      if (result==0) type[3] = ENCRYPTED;
   }
   catch(...)
   {
      status = FB_ENCRYPTEXC;
      type[3] = CLEARTEXT;
   }
   return result;
}
//---------------------------------------------------------------------------
int __fastcall TDataBuf::Decrypt(const AnsiString &password)
{
   status = DB_OK;
   int result;
   try
   {
      if (type[3]!=ENCRYPTED) return 0;
      result=data_decrypt(&data,length,password.c_str(),password.Length());
      if (result==0) type[3] = CLEARTEXT;
   }
   catch(...)
   {
      status = FB_DECRYPTEXC;
      type[3] = ENCRYPTED;
   }
   return result;
}
//---------------------------------------------------------------------------
bool __fastcall TDataBuf::LoadFromFile(const AnsiString &Filename)
{
   status = DB_OK;
   try
   {
      int len; Clear();
      if (!FileExists(Filename))
      {
         status = FB_NOFILEERR;
         return false;
      }
      if ((len=GetThisFileSize(Filename))==0)
      {
         Set(NULL,0);
         return true;
      }
      if ((data = new char[len+1])==NULL)
      {
         status = FB_ALLOCBUFERR;
         Set(NULL,0);
         return false;
      }
      TFileStream *f = new TFileStream(Filename,fmOpenRead);
      f->ReadBuffer(data,len); data[len] = '\0';
      delete f;
      orglength = len;
      length = len;
   }
   catch(xalloc)
   {
      status = FB_ALLOCBUFEXC;
      Set(NULL,0);
      return false;
   }
   catch(...)
   {
      status = FB_OPENREADFILEEXC; // ignore f
      Set(NULL,0);
      return false;
   }
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TDataBuf::SaveToFile(const AnsiString &Filename)
{
   status = DB_OK;
   try
   {
      TFileStream *f = new TFileStream(Filename,fmCreate);
      f->WriteBuffer(data,length);
      delete f;
   }
   catch(...)
   {
      status = FB_OPENWRITEFILEEXC; // ignore f
      return false;
   }
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TDataBuf::Compressed()
{
   return type[3]==COMPRESSED;
}
//---------------------------------------------------------------------------
int __fastcall TDataBuf::Compress()
{
   status = DB_OK;
   z_stream s;
   int err=0;

   s.zalloc = (alloc_func) 0;
   s.zfree = (free_func) 0;
   s.opaque = (voidpf) 0;
   s.next_in = NULL;

   try
   {
      if ((err=deflateInit2(&s,Z_DEFAULT_COMPRESSION,Z_DEFLATED,15,9,Z_DEFAULT_STRATEGY))==Z_OK)
      {
         int len = length*1.2+16;
         byte *outdata = new byte[len];

         orglength = length;  // should BE length (not 0) but it sometimes isn't...
         s.avail_in = length;
         s.next_in = (byte*)data;
         s.next_out = outdata;
         s.avail_out = len;
         if ((err=deflate(&s, Z_FINISH))==Z_STREAM_END)
         {
            if ((err=deflateEnd(&s))==Z_OK)
            {
               SetData((char*)outdata,s.total_out);
               type[3]=COMPRESSED;
            }
         }
         delete [] outdata;
      }
      if (err) status = FB_ZIPDEFLERR;
   }
   catch(...)
   {
      status = FB_ZIPDEFLEXC;
      if (!err) return Z_STREAM_ERROR;
   }
   return err;
}
//---------------------------------------------------------------------------
int __fastcall TDataBuf::Decompress()
{
   status = DB_OK;
   z_stream s;
   int err=0;

   s.zalloc = (alloc_func) 0;
   s.zfree = (free_func) 0;
   s.opaque = (voidpf) 0;

   s.next_in = (byte*)data;
   s.avail_in = length;
   s.next_out = NULL;

   try
   {
      if ((err=inflateInit2(&s,15))==Z_OK)
      {
         int len = orglength;
         byte *outdata = new byte[len];
         s.avail_out = len;
         s.next_out = outdata;
         if ((err=inflate(&s, Z_FINISH))==Z_STREAM_END)
         {
            if ((err=inflateEnd(&s))==Z_OK)
            {
               SetData((char*)outdata,s.total_out);
               type[3]=CLEARTEXT;
            }
         }
         delete [] outdata;
      }
      if (err) status = DB_ZIPINFLERR;
   }
   catch(...)
   {
      status = FB_ZIPDEFLEXC;
      if (!err) return Z_STREAM_ERROR;
   }
   return err;
}
//---------------------------------------------------------------------------
bool __fastcall TDataBuf::IsEmpty()
{
   return length==0;
}
//---------------------------------------------------------------------------
bool __fastcall TDataBuf::IsIndex()
{
   return type[1]==ISINDEX;
}
//---------------------------------------------------------------------------
bool __fastcall TDataBuf::IsData()
{
   return type[1]==ISDATA;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TDataBuf::str()
{
   return data ? AnsiString(data) : AnsiString("");
}
//---------------------------------------------------------------------------
char* __fastcall TDataBuf::c_str()
{
   char *doto = "\0";
   return data ? data : doto;
}
//---------------------------------------------------------------------------
// TFreeList
//---------------------------------------------------------------------------
__fastcall TFreeList::TFreeList() : TObject()
{
   status = FB_OK;
   offset = new TList();
   space = new TList();
   lastOffset = 0;
   lastSpace = 0;
}
//---------------------------------------------------------------------------
__fastcall TFreeList::~TFreeList()
{
   delete offset;
   delete space;
}
//---------------------------------------------------------------------------
bool __fastcall TFreeList::AddFreeSpace(TFLoffset Offset, TFLspace Space)
{
   int i=-1; while(++i<offset->Count)
      if (Offset<getOffset(i))
         break;

   offset->Insert(i,(void*)Offset); lastOffset = Offset;
   space->Insert(i,(void*)Space);  lastSpace = Space;
   return true;
}
//---------------------------------------------------------------------------
TFLoffset __fastcall TFreeList::Allocate(TFLspace Space)
{
   int i=-1; while (++i<offset->Count)
   {
      if (Space<=getSpace(i))
      {
         TFLoffset ofs = getOffset(i);
         space->Items[i] = (void*)(getSpace(i)-Space);
         if (space->Items[i] == 0)
         {
            offset->Delete(i);
            space->Delete(i);
         }
         else
            offset->Items[i] = (void*)(getOffset(i)+Space);
         return ofs;
      }
   }
   return FLNIL;
}
//---------------------------------------------------------------------------
void __fastcall TFreeList::AttemptMerge(int i)
{
   if (i<offset->Count-1)
   {
      if (getOffset(i)+(TFLoffset)getSpace(i)==getOffset(i+1))
      {
         space->Items[i] = (void*)(getSpace(i)+getSpace(i+1));
         offset->Delete(i+1);
         space->Delete(i+1);
      }
   }
}
//---------------------------------------------------------------------------
bool __fastcall TFreeList::Coalesce(TFLoffset Offset, TFLspace Space)
{
   int i=-1; while(++i<offset->Count)
   {
      // try merging to right of existing space...
      if (Offset == getOffset(i)+(TFLoffset)getSpace(i))
      {
         space->Items[i] = (void*)(getSpace(i)+Space);
         AttemptMerge(i);
         return true;
      }
      // try merging to left...
      if (Offset+Space == getOffset(i))
      {
         offset->Items[i] = (void*)Offset; lastOffset = Offset;
         space->Items[i] = (void*)(getSpace(i)+Space); lastSpace = getSpace(i);
         return true;
      }
   }
   return false;
}
//---------------------------------------------------------------------------
bool __fastcall TFreeList::EnterFreeSpace(TFLoffset Offset, TFLspace Space)
{
   if (Coalesce(Offset,Space)) return true;
   return AddFreeSpace(Offset,Space);
}
//---------------------------------------------------------------------------
bool __fastcall TFreeList::LoadFromStream(TStream *f)
{
  status = FL_OK;
  TFLspace spc; TFLoffset ofs;
  int i,count,spsize=SPCSIZE,ofsize=OFSSIZE,insize=sizeof(int);
  try
  {
     f->Read(&count,insize);
     if (offset->Count) offset->Clear();

     i=-1; while(++i<count)
     {
        f->Read(&ofs,ofsize);
        offset->Add((void*)ofs);
     }
     if (space->Count) space->Clear();
     i=-1; while(++i<count)
     {
        f->Read(&spc,spsize);
        space->Add((void*)spc);
     }
  }
  catch(...)
  {
     status = FL_READEXC;
     return false;
  }
  return true;
}
//---------------------------------------------------------------------------
bool __fastcall TFreeList::SaveToStream(TStream *f)
{
   status = FB_OK;
   TFLspace spc; TFLoffset ofs;
   int i,count=offset->Count,ofsize=OFSSIZE,spsize=SPCSIZE,insize=sizeof(int);
   try
   {
      f->Write(&count,insize);
      i=-1; while(++i<count)
      {
         ofs = getOffset(i);
         f->Write(&ofs,ofsize);
      }
      i=-1; while(++i<count)
      {
         spc = getSpace(i);
         f->Write(&spc,spsize);
      }
   }
   catch(...)
   {
      status = FL_WRITEEXC;
      return false;
   }
   return true;
}
//---------------------------------------------------------------------------
int __fastcall TFreeList::Count()
{
   return offset->Count;
}
//---------------------------------------------------------------------------
// TFMindex
//---------------------------------------------------------------------------
__fastcall TFMindex::TFMindex(TFileManager *Fm, AnsiString Name, TFMindex *Parent)
{
   fm = Fm;
   name = Name;
   parent = Parent;
   try
   {
   if (parent && !name.IsEmpty()) offset = parent->GetOffset(name);
   else
   {
      // initialize first allocation if necessary
      if (fm->start()==FLNIL)
      {
         offset = FLNIL;
         if (!fm->WriteOfs(fm->Allocate(OFSSIZE),&offset,OFSSIZE))
         {
            fm->status(DB_WRITEINDEXOFSERR);
            return;
         }
      }
      else // or just read offset
      {
         if (!fm->ReadOfs(fm->start(),&offset,OFSSIZE))
         {
            fm->status(DB_READINDEXERR);
            return;
         }
      }
   }
   if (offset==FLNIL) return;

   void *p;
   TDataBuf *buf = new TDataBuf();
   if (buf)
   {
      if (!fm->ReadOfs(offset,&buf->type[0],TYPSIZE) ||
          !fm->ReadDat(&buf->length,SPCSIZE) ||
          !fm->ReadDat(&buf->orglength,SPCSIZE))
      {
         fm->status(DB_READINDEXERR);
         delete buf;
         return;
      }

      if (buf->data) delete [] buf->data;
      buf->data = new char[buf->length];

      if (!fm->ReadDat(buf->data,buf->length))
      {
         fm->status(DB_READINDEXERR);
         delete buf;
         return;
      }
      if (buf->Compressed() && buf->Decompress()!=DB_OK)
      {
         fm->status(DB_ZIPINFLERR);
         fm->status(DB_READINDEXERR);
         delete buf;
         return;
      }
      Text = AnsiString(buf->data);
      int i=-1; while(++i<Count)
      {
         if (!fm->ReadDat(&p,PTRSIZE))
         {
            fm->status(DB_READINDEXERR);
            delete buf;
            return;
         }
         Objects[i] = (TObject*)p;
      }
      delete buf;
   }
   }
   catch(xalloc)
   {
      fm->status(DB_ALLOCINDEXEXC);
   }
   catch(...)
   {
      fm->status(DB_READINDEXEXC);
   }
}
//---------------------------------------------------------------------------
void __fastcall TFMindex::ShowIndex()
{
   String idx;
   for (int i=0; i<Count; i++)
      idx += Strings[i]+"     "+String((int)Objects[i])+"\n";
   ShowMessage(idx);
}
//---------------------------------------------------------------------------
TFLspace __fastcall TFMindex::GetSize(AnsiString &key)
{
   TFLoffset ofs = GetOffset(key);
   if (ofs==FLNIL) return 0;

   TFLspace spc=0;
   char type[TYPSIZE];
   if (!fm->ReadOfs(ofs,&type[0],TYPSIZE) ||
       !fm->ReadDat(&spc,SPCSIZE))
   {
      fm->status(DB_GETDATSIZEERR);
      return 0;
   }
   return spc;
}
//---------------------------------------------------------------------------
char __fastcall TFMindex::GetType(AnsiString &key)
{
   TFLoffset ofs = GetOffset(key);
   if (ofs==FLNIL) return -1;

   char type[TYPSIZE];
   if (!fm->ReadOfs(ofs,&type[0],TYPSIZE))
   {
      fm->status(DB_GETDATTYPEERR);
      return -1;
   }
   return type[0];
}
//---------------------------------------------------------------------------
TFLoffset __fastcall TFMindex::GetOffset(AnsiString &key)
{
   int p = IndexOf(key);
   return p<0 ? FLNIL : (TFLoffset)Objects[p];
}
//---------------------------------------------------------------------------
TFLoffset __fastcall TFMindex::SetOffset(AnsiString &key, TFLoffset Offset)
{
   if (key.IsEmpty()) return FLNIL;
   int p = IndexOf(key);
   if (p < 0)
   {
      Add(key);
      p = Count-1;
   }
   Objects[p] = (TObject*)Offset;
   return Offset;
}
//---------------------------------------------------------------------------
TFLoffset __fastcall TFMindex::PutOffset(AnsiString &key, TFLoffset Offset)
{
   TFLoffset ofs = SetOffset(key,Offset); Save();
   return ofs;
}
//---------------------------------------------------------------------------
TFLoffset __fastcall TFMindex::WriteIndex()
{
   if (offset != FLNIL) // deallocate previous offset to index
      fm->Deallocate(offset);

   TDataBuf *buf = new TDataBuf(Text.c_str(),Text.Length()+1);
   if (buf)
   {
      buf->type[0] = ISINDEX;
      buf->type[1] = 0;
      buf->type[2] = 0;
      buf->type[3] = 0;
      if (UseCompression && buf->Compress()!=FB_OK)
      {
         fm->status(DB_ZIPDEFLERR);
         fm->status(DB_WRITEINDEXERR);
         delete buf;
         return FLNIL;
      }
      offset = fm->Allocate(TYPSIZE+(SPCSIZE*2)+buf->length+(Count*PTRSIZE));
      if (!fm->WriteOfs(offset,&buf->type[0],TYPSIZE) ||
          !fm->WriteDat(&buf->length,SPCSIZE) ||
          !fm->WriteDat(&buf->orglength,SPCSIZE) ||
          !fm->WriteDat(buf->data,buf->length))
      {
         fm->status(DB_WRITEINDEXERR);
         delete buf;
         return FLNIL;
      }
      delete buf;
      int i=-1; while(++i<Count)
      {
         void *p = Objects[i];
         if (!fm->WriteDat(&p,PTRSIZE))
         {
            fm->status(DB_WRITEINDEXERR);
            return FLNIL;
         }
      }
   }
   return offset;
}
//---------------------------------------------------------------------------
TFLoffset __fastcall TFMindex::Save()
{
   TFLoffset ofs = WriteIndex();
   if (parent) parent->PutOffset(name,offset);
   else
   {
      if (!fm->WriteOfs(fm->start(),&offset,OFSSIZE))
      {
         fm->status(DB_SAVEINDEXERR);
         return FLNIL;
      }
   }
   return ofs;
}
//---------------------------------------------------------------------------
char* __fastcall TFMindex::GetBinary(AnsiString &key, TDataBuf *buf)
{
   if (!buf) return NULL;
   TFLoffset ofs = GetOffset(key);
   if (ofs == FLNIL) return NULL;

   try
   {
      if (!fm->ReadOfs(ofs,&buf->type[0],TYPSIZE) ||
          !fm->ReadDat(&buf->length,SPCSIZE) ||
          !fm->ReadDat(&buf->orglength,SPCSIZE))
      {
         fm->status(DB_GETBINARYERR);
         buf->Clear();
         return NULL;
      }

      if (buf->data) delete[] buf->data;
      buf->data = new char[buf->length+1];
      buf->data[buf->length]=0;

      if (!fm->ReadDat(buf->data,buf->length))
      {
         fm->status(DB_GETBINARYERR);
         buf->Clear();
         return NULL;
      }
      if (buf->Compressed() && buf->Decompress()!=FB_OK)
      {
         fm->status(DB_ZIPINFLERR);
         fm->status(DB_GETBINARYERR);
         buf->Clear();
         return NULL;
      }
   }
   catch (xalloc)
   {
      fm->status(DB_GETBINALLOCEXC);
      buf->Clear();
      return NULL;
   }
   return buf->data;
}
//---------------------------------------------------------------------------
/*TFLoffset __fastcall TFMindex::PutBinary(AnsiString &key, TDataBuf* buf, bool SaveIndex)
{
   // get key id or add if not already there
   int pos = IndexOf(key);
   if (pos < 0)
   {
      AddObject(key,(TObject*)FLNIL);
      pos = Count-1;
   }
   // get offset...deallocate, allocate...write space then data
   TFLoffset ofs = GetOffset(key);
   if (ofs != FLNIL)
      fm->Deallocate(ofs);

   if (!buf)
      Delete(IndexOf(key));
   else
   {
      void *p;
      buf->type[0] = ISDATA;
      buf->type[1] = 0;
      buf->type[2] = 0;
      buf->type[3] = 0;
      if (UseCompression) buf->Compress();
      if ((ofs = fm->Allocate(TYPSIZE+(SPCSIZE*2)+buf->length+(Count*PTRSIZE))) != FLNIL)
      {
         fm->SeekTo(ofs);
         fm->WriteDat(&buf->type[0],TYPSIZE);
         fm->WriteDat(&buf->length,SPCSIZE);
         fm->WriteDat(&buf->orglength,SPCSIZE);
         fm->WriteDat(buf->data,buf->length);
         // added to make index and data saves sameo sameo
         for (int i=0; i<Count; i++)
         {
            p = Objects[i];
            fm->WriteDat(&p,PTRSIZE);
         }
      }
      // assign new offset to index key and save
      Objects[pos] = (TObject*)ofs;
   }
   if (SaveIndex) Save();
   return ofs;
}*/
//---------------------------------------------------------------------------
TFLoffset __fastcall TFMindex::PutBinary(AnsiString &key, TDataBuf* buf, bool SaveIndex)
{
   // get key id or add if not already there
   int pos = IndexOf(key);
   if (pos < 0)
   {
      AddObject(key,(TObject*)FLNIL);
      pos = Count-1;
   }
   // get offset...deallocate, allocate...write space then data
   TFLoffset ofs=GetOffset(key);
   if (ofs!=FLNIL) fm->Deallocate(ofs);
   if (!buf) Delete(IndexOf(key));
   else
   {
      void *p;
      buf->type[0] = ISDATA;
      buf->type[1] = 0;
      buf->type[2] = 0;
      buf->type[3] = 0;
      if (UseCompression && buf->Compress()!=FB_OK)
      {
         fm->status(DB_ZIPDEFLERR);
         fm->status(DB_PUTBINARYERR);
         return ofs;
      }
      if ((ofs=fm->Allocate(TYPSIZE+(SPCSIZE*2)+buf->length+(Count*PTRSIZE)))!=FLNIL)
      {
         if (!fm->WriteOfs(ofs,&buf->type[0],TYPSIZE) ||
             !fm->WriteDat(&buf->length,SPCSIZE) ||
             !fm->WriteDat(&buf->orglength,SPCSIZE) ||
             !fm->WriteDat(buf->data,buf->length))
         {
            fm->status(DB_PUTBINARYERR);
            return ofs;
         }
         // added to make index and data saves sameo sameo
         for (int i=0; i<Count; i++)
         {
            p = Objects[i];
            if (!fm->WriteDat(&p,PTRSIZE))
            {
               fm->status(DB_PUTBINARYERR);
               return ofs;
            }
         }
      }
      // assign new offset to index key and save
      Objects[pos] = (TObject*)ofs;
   }
   if (SaveIndex) Save();
   return ofs;
}
//---------------------------------------------------------------------------
/*int __fastcall PutAppBinary(AnsiString &path, TMemoryStream *buf, AnsiString &datafile)
{
   String nodetext;
   int status = DB_OK;
   TFMindex *idx[16] = { 0 };
   TFileManager *fman = new TFileManager(datafile);
   idx[0] = new TFMindex(fman);

   int i=1; while ((nodetext = GetParam(i-1,path,PS.c_str())) != NULL)
   {
      if (GetParam(i,path,PS.c_str()).IsEmpty())
      {
         idx[i-1]->PutBinary(nodetext,buf);
         status = fman->status;
         break;
      }
      idx[i] = new TFMindex(fman,nodetext,idx[i-1]);
      i++;
   }

   for (i=i-1; i>0; i--)
      delete idx[i];
   delete fman;
   return status;
}*/
//---------------------------------------------------------------------------




