//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "MimeEncDlgs.h"
#include "MimeMan.h"
#include "FileMan.h"
#include "global.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TMimeEncDlg *MimeEncDlg;
TMimeEnc *MimeEnc=NULL;
//---------------------------------------------------------------------------
bool __fastcall AttachFile(TForm *Owner, TStrings *Output)
{
   MimeEncDlg = new TMimeEncDlg(Owner);
   bool ok = MimeEncDlg->Execute(Output);
   delete MimeEncDlg;
   return ok;
}
//---------------------------------------------------------------------------
__fastcall TMimeEncDlg::TMimeEncDlg(TComponent* Owner)
	: TForm(Owner)
{
   CenterDialog(this,Owner);
   OpenDialog->InitialDir = AppDir;
   Edit1->Text = GetAppIni("Last Attachment","");
}
//---------------------------------------------------------------------------
bool __fastcall TMimeEncDlg::Execute(TStrings *Output)
{
   output = Output;
   ShowModal();
   return status==statMimeOk;
}
//---------------------------------------------------------------------------
void __fastcall TMimeEncDlg::AttachBtnClick(TObject *Sender)
{
   AnsiString filename = Edit1->Text;
   if (filename.IsEmpty())
   {
      MessageBeep(0);
      return;
   }
   if (!FileExists(filename))
   {
      MsgDlg(this,"The file you specified does not exist!","Error",MB_OK);
      return;
   }
   if (MsgDlg(this,"Attach '"+filename+"' now?","Confirm",MB_OKCANCEL)==IDCANCEL)
      return;

   MimeEnc = new TMimeEnc(filename,output,ProgressBar);
   status = MimeEnc->Execute();
   delete MimeEnc;
   MimeEnc = NULL;
   if (status==statMimeOk)
   {
      PutAppIni("Last Attachment",Edit1->Text);
      ModalResult = mrOk;
      Snooze(200);
   }
}
//---------------------------------------------------------------------------
void __fastcall TMimeEncDlg::BrowseBtnClick(TObject *Sender)
{
   if (OpenDialog->Execute()) Edit1->Text = OpenDialog->FileName;
}
//---------------------------------------------------------------------------
__fastcall TMimeEnc::TMimeEnc(AnsiString filename, TStrings *output, TProgressBar *progressbar)
{
   Filename = filename;
   ProgressBar = progressbar;
   ProgressBar->Position = 0;
   ProgressBar->Max = 100;
   ProgressBar->Min = 0;
   status = statMimeOk;
   Output = output;
   Cancelled = false;
}
//---------------------------------------------------------------------------
__fastcall TMimeEnc::~TMimeEnc()
{
}
//---------------------------------------------------------------------------
int __fastcall TMimeEnc::Execute()
{
   AnsiString FileName = ExtractFileName(Filename);
   Output->Add("Content-Type: "+GetContentType(FileName)+"; name=\""+FileName+"\"");
   Output->Add("Content-Transfer-Encoding: base64");
   Output->Add("Content-Disposition: attachment; filename=\""+FileName+"\"");
   Output->Add("");
   UUEncode();
   return status;
}
//---------------------------------------------------------------------------
void __fastcall TMimeEnc::UUEncode()
{
   const char Base64Chars[] =
   {
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
    'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
    'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/', '='
   };

   char DataIn[3];
   char DataOut[81];
   ProgressBar->Position = 0;
   ProgressBar->Visible = true;
   int i,Count=0,BytesRd,InBytes=sizeof(DataIn);
   TFileStream *f = new TFileStream(Filename,fmOpenRead);
   if (!f) return;

   while (true)
   {
      BytesRd=f->Read(DataIn,InBytes);
      for (i=1; i<4; i++) DataOut[Count+i]=0x40;

      double p=f->Position,s=f->Size,pos=(p/s)*100;
      ProgressBar->Position = (WORD)pos; Application->ProcessMessages();

      DataOut[Count]   = char((DataIn[0] & 0xFC) >> 2);
      DataOut[Count+1] = char((DataIn[0] & 0x03) << 4);
      if (BytesRd>1)
      {
         DataOut[Count+1] += char((DataIn[1] & 0xF0) >> 4);
         DataOut[Count+2]  = char((DataIn[1] & 0x0F) << 2);
         if (BytesRd>2)
         {
            DataOut[Count+2] += char((DataIn[2] & 0xC0) >> 6);
            DataOut[Count+3]  =  char(DataIn[2] & 0x3F);
         }
      }

      for (i=0; i<4; i++) DataOut[Count+i] = byte(Base64Chars[DataOut[Count+i]]);

      Count += 4; if (Count<60 && BytesRd==InBytes) continue;

      DataOut[Count] = '\0';
      Output->Add(AnsiString(&DataOut[0]));

      if (Cancelled) {
         status = statMimeCancel;
         break;
      }

      if (BytesRd<InBytes) break;

      Count=0;
   }
   ProgressBar->Visible = false;
   delete f;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMimeEnc::GetContentType(AnsiString &Filename)
{
  AnsiString type,Ext = ExtractFileExt(Filename).UpperCase();
  if (Ext==".HTM" || Ext==".HTML") type = "text/html"; else
  if (Ext==".AU"  || Ext==".SND" ) type = "audio/basic"; else
  if (Ext==".AIF") type = "audio/aiff"; else
  if (Ext==".GIF") type = "image/gif"; else
  if (Ext==".JPG") type = "image/jpeg"; else
  if (Ext==".AVI") type = "video/avi"; else
  if (Ext==".RTF") type = "text/rtf"; else
  if (Ext==".TXT") type = "text/plain"; else
     type = "application/octet-stream";
  return type;
}
//---------------------------------------------------------------------------
void __fastcall TMimeEncDlg::CancelBtnClick(TObject *Sender)
{
   if (MimeEnc)
   {
      MimeEnc->Cancelled = true;
      MsgDlg(this,"MIME encoding operation cancelled!","Information",MB_OK);
   }
   else ModalResult = mrCancel;
}
//---------------------------------------------------------------------------
