//---------------------------------------------------------------------------
#ifndef PGPbuffersH
#define PGPbuffersH

#include "mixmaster.h"
#include "fileman.h"
#include "TBUFFER.h"
//---------------------------------------------------------------------------
#define ENABLE true
#define REWIND true

class TPGPbuffer : public TBUFFER
{
public:
   // properties
   __property byte CTB                          = { read=GetCTB };
   __property byte version                      = { read=fversion };   //GetVersion };
   __property int timestamp                     = { read=funixtime };  //GetTimeStamp };
   __property int validity                      = { read=fvalid };     //GetValidity };
   __property AnsiString idtext8                = { read=GetIdText8 };
   __property AnsiString firstuid               = { read=FirstUID };
   __property int timestr                       = { read=TimeStr };
   __property int bits                          = { read=fmodbits };
   __property int expbits                       = { read=fexpbits };
   __property byte algo                         = { read=falgo };
   __property AnsiString algostr                = { read=GetAlgoStr };
   __property AnsiString bitsstr                = { read=GetBitsStr };
   __property bool ispair                       = { read=IsPair };
   __property bool ispublic                     = { read=IsPublic };
   __property bool issecret                     = { read=IsSecret };
   __property int count                         = { read=Count };
   __property bool enabled                      = { read=Enabled };
   __property bool revoked                      = { read=Revoked };
   __property bool expired                      = { read=Expired };
   __property bool eok                          = { read=EndOfKey };
   __property bool parsed                       = { read=fparsed };

   virtual __fastcall TPGPbuffer();
   virtual __fastcall TPGPbuffer(AnsiString keyringfile);
   virtual __fastcall TPGPbuffer(AnsiString keyringfile, bool includedisables);
   virtual __fastcall TPGPbuffer(TStrings *asciiblock);
   virtual __fastcall TPGPbuffer(TPGPbuffer *buf);
   virtual __fastcall~TPGPbuffer();
   void __fastcall import(TStrings *asciiblock);
   void __fastcall import(const AnsiString &asciiblock);
   void __fastcall export(TStrings *asciiblock);
   AnsiString __fastcall export();
   void __fastcall exporttrust(TStrings *asciiblock);
   void __fastcall loadfromfile(AnsiString keyringfile);
   void __fastcall loadfromfileex(AnsiString keyringfile);
   bool __fastcall copyfrom(TPGPbuffer *buffer, int length=0);
   bool __fastcall copypacketfrom(TPGPbuffer *pkts, bool rewind=false);
   bool __fastcall copykeyfrom(TPGPbuffer *keys, bool rewind=false); // sets ptr to 0 when done, ready to process key
   bool __fastcall appendkeyfrom(TPGPbuffer *keys, bool rewind=false);
   bool __fastcall steppacket(); 
   bool __fastcall hasuserid(AnsiString &id);
   int  __fastcall onlyusefulkeys(); // removes disables/revokes/expires
   void __fastcall getuserids(TStrings *ids);
   AnsiString __fastcall getuserids();
   void __fastcall getdescrips(TStrings *lines);
   AnsiString __fastcall getdescrips();
   bool __fastcall keyexists(TPGPbuffer *key);
   bool __fastcall get(AnsiString IdText8, TPGPbuffer *key);
   AnsiString __fastcall getidtext8(AnsiString UserId);
   int  __fastcall keyid(TPGPbuffer *keyid);
   int  __fastcall keyidex(BUFFER *keyid);
   bool __fastcall parsekeycert(bool force=false);
   bool __fastcall enable(bool enable);

   void __fastcall cpkenable(bool enable);
   void __fastcall cpkaddtrust();

   void __fastcall disable();
   void __fastcall revoke();
   void __fastcall clear();
   bool __fastcall save();

   byte __fastcall GetCTB();
   byte __fastcall GetVersion();
   int  __fastcall GetTimeStamp();
   int  __fastcall GetValidity();

   bool __fastcall EndOfKey(); // is the buf ptr on an eok byte
   int  __fastcall GetPacketSize(); // whole pktlen, ptr at CTB
   int  __fastcall GetPacketInfo(byte *CTB, byte *lolf=NULL);// datlen & CTB, ptr at data
   
   virtual byte __fastcall ReadByte();
   virtual int  __fastcall ReadInt(); // reads 2 bytes not 4
   virtual long __fastcall ReadLong();
   virtual int  __fastcall WriteByte(byte b);
   virtual int  __fastcall WriteInt(int i); // writes 2 bytes not 4
   virtual int  __fastcall WriteLong(long l);
   AnsiString __fastcall PKAlgoStr(int Algo);
   AnsiString __fastcall GetAlgoStr();
   AnsiString __fastcall GetBitsStr();
   AnsiString __fastcall FirstUID();
   AnsiString __fastcall GetIdText8();
   AnsiString __fastcall IdText8();
   bool __fastcall IsPublic();
   bool __fastcall IsSecret();
   bool __fastcall Enabled();
   bool __fastcall Revoked();
   bool __fastcall Expired();
   bool __fastcall IsPair();
   byte __fastcall Version();
   int  __fastcall Count();
   AnsiString __fastcall TimeStr();
private:
   AnsiString fidtext8;
   byte fversion;
   int funixtime;
   int fvalid;
   byte falgo;
   int fmodbits;
   int fexpbits;
   int falgocip;
   int fdexbits;
   int ffacpbits;
   int ffacqbits;
   int finvbits;
   int fchecksum;
   byte fsubalgo;
   int fsubbits;
   bool fparsed;
};

typedef TPGPbuffer TPGPkey;

class TPGPkeylist : public TStringList
{
public:
   virtual __fastcall TPGPkeylist();
   virtual __fastcall TPGPkeylist(AnsiString Filename);
   virtual __fastcall TPGPkeylist(TPGPbuffer *keys);
   virtual __fastcall~TPGPkeylist();
   virtual TPGPbuffer*  __fastcall PgpKey(int atPos);
   virtual AnsiString __fastcall KeyDescription(int atPos);
   virtual int  __fastcall AddKeys(TPGPkeylist *keys);
   virtual int  __fastcall AddKeys(TPGPbuffer *keys);
   virtual bool __fastcall DeleteKey(int atPos);
   virtual int  __fastcall IndexOf(AnsiString idtext8);
   virtual void __fastcall Clear();
};

#define NO_DISABLED false

#define PgpButtonsHdr "Pgp-Buttons: "
#define PgpHdr "Pgp: "

// packet types
#define PublicKeyPacket(CTB)    (CTB==153)
#define PublicSubKeyPacket(CTB) (CTB==185)
#define SecretKeyPacket(CTB)    (CTB==149)
#define SignaturePacket(CTB)    (CTB==137)
#define UserIdPacket(CTB)       (CTB==180)
#define TrustPacket(CTB)        (CTB==176)

// algorithms
#define PGP_ANY 0
#define PGP_ES_RSA 1
#define PGP_E_ELG 16
#define PGP_S_DSA 17

// mix's pgp defines
#define MSG_MAIL 1
#define PGP_PASS -2
#define PGP_ERR -1
#define PGP_SIGN 1
#define PGP_ENCRYPT 2
#define PGP_TEXT 16
#define PGP_REMAIL 8
#define PGP_CONVENTIONAL 4
#define PGP_NOARMOR 32

// mix's pgp error codes
#define PGP_OK 0	 	    // valid message, not signed
#define PGP_SIGOK 1	    // valid signature
#define PGP_NOMSG 2		 // is not a OpenPGP message
#define PGP_NODATA 3	 // OpenPGP packet does not contain user data
#define PGP_SIGNKEY 4	 // can't verify signature
#define PGP_ERR -1		 // can't read message, no matching key found
#define PGP_PASS -2		 // bad passphrase
#define PGP_SIGBAD -3	 // bad signature

#define PGP_DSKCRUNCH -16 // not enough disk space to save encrypted file
// packet type
#define PGP_ENCRYPTED 9

// SHA in libeay.dll
int __fastcall digestmem_sha1ex(byte *b, int n, BUFFER *md);
int __fastcall digest_sha1ex(BUFFER *b, BUFFER *md);

// ssllib crypto
#define SHA_DIGEST_LENGTH 20
void __fastcall SHA1ex (byte *b, int n, byte* m);
typedef void (*_SHA1ex)(byte*, int, byte*);

// change a byte value to string of 1s and 0s
AnsiString __fastcall BinaryByte(byte value, byte frombit=7, byte tobit=0);

// mixlib pgp functions
int __fastcall pgpencrypt(int mode, BUFFER *message, BUFFER *encr, BUFFER *sigid,
		BUFFER *pass, char *pubring, char *secring);
typedef int (*PGPENCRYPT)(int, BUFFER*, BUFFER*, BUFFER*, BUFFER*, char*, char*);

// misc pgp support functions
int __fastcall pgparmor(BUFFER *in, int mode);
int __fastcall pgpdearmor(BUFFER *in, BUFFER *out);
int __fastcall pgpgetsize(BUFFER *packet, byte ctb=0);
int __fastcall pgppackettype(BUFFER *b, long *len);
int __fastcall pgpispacket(BUFFER *b);
int __fastcall pgpisconventional(BUFFER *b);
// mpi functions
int __fastcall mpinum(int algo);
int __fastcall mpiput(BUFFER *buf, BUFFER *mpi);
int __fastcall mpiget(BUFFER *b, BUFFER *mpi);

bool __fastcall ParsePublicKeyPacket(BUFFER *packet, AnsiString *idtext8, int *version,
   int *unixtime, int *valid, int *algo, int *modbits, int *expbits, BUFFER *modulus,
   BUFFER *exponent);

bool __fastcall ParseSecretKeyPacket(BUFFER *packet,  AnsiString *idtext8, byte *version,
   int *unixtime, int *valid, byte *algo, int *modbits, int *expbits, BUFFER *modulus,
   BUFFER *exponent, int *algocip, BUFFER *cipfbIV, BUFFER *dexponent, int *dexbits,
   BUFFER *factorp, int *facpbits, BUFFER *factorq, int *facqbits,
   BUFFER *inverse, int *invbits, int *checksum);

bool __fastcall ParseSignaturePacket3(BUFFER *packet, int *version, int *implicit,
   int *sigclass, int *unixtime, AnsiString *idtext8, int *pkalgo,
   int *mdalgo, int *chksum);

bool __fastcall ParseSignaturePacket4(BUFFER *packet, int *version,
   int *sigclass, int *hashalgo, int *pkalgo,
   AnsiString *idtext8, int *nestflag);

bool __fastcall ParseUserIdPacket(BUFFER *packet, AnsiString &userid);

#define begin_pgp "-----BEGIN PGP "
#define begin_pgpmsg "-----BEGIN PGP MESSAGE-----"
#define begin_pgpkey "-----BEGIN PGP PUBLIC KEY BLOCK-----"
#define begin_pgpprivatekey "-----BEGIN PGP PRIVATE KEY BLOCK-----"
#define begin_pgpsigned "-----BEGIN PGP SIGNED MESSAGE-----"
#define begin_pgpsig "-----BEGIN PGP SIGNATURE-----"
#define end_pgp "-----END PGP "
#define end_pgpmsg "-----END PGP MESSAGE-----"
#define end_pgpkey "-----END PGP PUBLIC KEY BLOCK-----"
#define end_pgpprivatekey "-----END PGP PRIVATE KEY BLOCK-----"
#define end_pgpsig "-----END PGP SIGNATURE-----"

#endif
